//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      Clans.java                                                       *
//* CLASS    Ecology.java. (a grid-based, n x m synthetic ecology)            *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     07/01/2015                                                       *
//* REVISION X.0.1.4                                                          *
//* BYTES    00,000                                                           *
//* TIME     6:30                                                             *
//* PURPOSE  This object binds together the simulated ecology.                *
//* METHOD   Each instance of a class is a stand-alone code object.           *
//* DETAILS  A Java document exists that outlines class operation.  Other     *
//*          documentation in the form of programmer notes apply.             *
//* ENVIRON  Developed on Intel Core-2 Quad Q8200 2.33 GHz CPU, Nvidia-based  *
//*          PCI-X graphics acceleration.  The OS was MS Windows 7 Ultimate.  *
//*          Java IDE was JGrasp version 2.0.0_08.  The Java compiler was     *
//*          version 1.8.0.  Verification tests were run in Java JRE          *
//*          build 1.8.0-b132 with client VM build 14.3-b01.                  *
//*                                                                           *
//*****************************************************************************
//

package root.ecology;                                 // sub-directory storage.

import java.util.ArrayList;                                   // Java language.
import java.util.Random;                                      //         ditto.

import root.ecology.Cell;                                  // program routines.
import root.ecology.climate.Climate;                       //            ditto.
import root.ecology.fauna.Fauna;                           //            ditto.
import root.ecology.flora.Flora;                           //            ditto.
import root.ecology.groups.Groups;                         //            ditto.
import root.ecology.terrain.Terrain;                       //            ditto.
import root.hominini.Hominid;                              //            ditto.
import root.ModelConstraints;                              //            ditto.
import root.rng.TrueRNG;                                   //            ditto.

//
//*****************************************************************************
// Ecology

public class Ecology extends ModelConstraints
{

  //
  //***************************************************************************
  // Class constants, variables, and objects.

  // Constants.
  final boolean ODD = true;     // used to control the placement of foragables.
  final boolean EVEN = false;   // used to control the placement of foragables.

  final int CLIMATE_CHANGE_INTERVAL = YEAR;     // count of days before change.

  // Variables.
  private long rng_Seed = 0;                    // this is the GLOBAL RNG seed.
  private int horizontalLimit = 0;                  // horizontal grid surface.
  private int verticalLimit = 0;                      // vertical grid surface.
  private int galoisSetCardinality = 0;  // the m X n size of the grid surface.
  private static int [] urDistribution = null;            // a UR distribution.
  private int localEpoch = 0;               // local copy of the current epoch.

  // Objects.

  // a 2D object built of Cells having the properties of an ecology (like
  // temperature, water, forageable, etc.) and occupied by hominid cohorts.
  public static Cell [][] grid = null;

  public static Climate climate = null;  // rules, seasons, precipitation, etc.
  public static Fauna fauna = null;         // rules, foragable animal species.
  public static Flora flora = null;    // rules, foragable vegetable resources.
  public static Groups groups = null; //rules, hominid group locator resources.
  public static Terrain terrain = null; // tools used to manipulate the ground.

  private static Random RNG = null;          // pointer to standard RNG object.
  private static TrueRNG trueRNG = null;      // pointer to TrueRNG USB device.

  //***************************************************************************
  /** This is the Ecology class constructor.  It instantiates an matrix of
      software objects representing an abstract, 2-D surface.  The objects
      are one cell deep and are distributed across (and completely cover)
      an hLimit x vLimit rectangular grid.*/

  public Ecology (long newSeed, TrueRNG devicePntr, int hLimit, int vLimit)
  {

    this.rng_Seed = newSeed;            // make a local copy of the fixed seed.

    // if the TrueRNG USB device is available (ie devicePntr is not null) then,
    // the seed used by RNG will be truly random.  Otherwise, a fixed value of
    // newSeed will be used and all stochastic patterns will be deterministic.
    if (devicePntr == null)
      RNG = new Random(newSeed);  // instantiate an RNG object with fixed seed.
    else
    {
      trueRNG = devicePntr;                // make a local copy of the pointer.
      RNG = new Random (devicePntr.getNextRandom()); // create with real rndom.
    }

    this.horizontalLimit = hLimit;                        // take a local copy.
    this.verticalLimit = vLimit;                          // take a local copy.

    // USER MUST SELECT TO: generate a distribution using one of four methods.
    this.galoisSetCardinality = this.horizontalLimit * this.verticalLimit;
    // 1. UR distribution.
    System.out.println (" (UR distribution).");
    urDistribution = getUniformRandom(galoisSetCardinality);
    // 2. Java factory distribution.
    //System.out.println (" (Java factory distribution).");
    //urDistribution = factoryUR(galoisSetCardinality);
    // 3. bounded leptokurdic distribution.
    //System.out.println (" (bounded leptokurdic distribution).");
    //urDistribution = trueRNGUR(galoisSetCardinality);
    // 4. hybridized distribution.
    //System.out.println (" (hybridized distribution).");
    //urDistribution = hybridRNG(galoisSetCardinality);

    // grid is a 2D array of objects emulating the ecology's terrain basis.
    grid = new Cell[this.horizontalLimit][this.verticalLimit];
    grid = startTheGrid (grid);       // a VERY BASIC Cell-based instantiation.
    // OK, Ecology now owns the "grid."

    // terrain is the set of objects that comprise the grid.
    terrain = new Terrain (this.rng_Seed, grid,
                           this.horizontalLimit, this.verticalLimit);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and adds the
      individual hominid (also named by parameter) to the cell's cohort.
      It calls a method in Terrain to accomplish its goal.*/

  public void addCohortIndividual (int h, int v, Hominid thisHominid)
  {

    groups.addCohortIndividual (h, v, thisHominid);

  } // End method.
  //***************************************************************************


  //***************************************************************************
  /** This method directs the aging of the system ecology using independent
      threads - fauna, flora, and climate.*/

  public void ageEcoSystem (int epoch)
  {

    if ((epoch % CLIMATE_CHANGE_INTERVAL ) == 0)
    {
      Thread climateThread = null;                       // climate background.
      climate.setClimateGrid (grid);                  // pass the grid pointer.
      climateThread = new Thread (climate);              // climate background.
      climateThread.start ();                              // start the thread.
      try
      {
        climateThread.join ();                   // wait while thread finishes.
      }
      catch (InterruptedException ie)
      {
      }

      Thread terrainThread = null;                       // terrain background.
      terrain.setTerrainGrid (grid);                  // pass the grid pointer.
      terrainThread = new Thread (terrain);              // terrain background.
      terrainThread.start ();                              // start the thread.
      try
      {
        terrainThread.join ();                   // wait while thread finishes.
      }
      catch (InterruptedException ie)
      {
      }

      climateThread = null;       // return the memory to the operating system.
      terrainThread = null;       // return the memory to the operating system.
    }

    Thread faunaThread = null;                             // fauna background.
    fauna.setEpoch (epoch);
    fauna.setFaunaGrid (grid);                        // pass the grid pointer.
    faunaThread = new Thread (fauna);                      // fauna background.
    faunaThread.start ();                                  // start the thread.
    try
    {
      faunaThread.join ();                       // wait while thread finishes.
    }
    catch (InterruptedException ie)
    {
    }

    Thread floraThread = null;                             // flora background.
    flora.setEpoch (epoch);
    floraThread = new Thread (flora);                      // flora background.
    floraThread.start ();                                  // start the thread.
    try
    {
      floraThread.join ();                       // wait while thread finishes.
    }
    catch (InterruptedException ie)
    {
    }

    faunaThread = null;           // return the memory to the operating system.
    floraThread = null;           // return the memory to the operating system.

    // DIAGNOSTIC
    //int x = 37;
    //int y = 52;
    //int x = 325;
    //int y = 350;
    //float prob = getForageProbability (x, y);
    //float prod = getMaxProductivity (x, y);
    //float water = getBiomeWater (x, y);
    //System.out.println("Ecology.ageEcoSystem x=" + x + ", y= " + y +
    //                   " prob=" + prob + " prod=" + prod +
    //                   " water= " + water);
    // DIAGNOSTIC

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method calculates and returns the average cell cohort occupancy.*/

  public int averageCohortSize ()
  {

    return groups.averageCohortSize ();

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns an array of integers built using the factory RNG.*/

  public int [] factoryUR (int galoisFieldSize)
  {

    int array [] = new int[galoisFieldSize];               // the return field.

    for (int i = 0; i < galoisFieldSize; i++)
      array[i] = RNG.nextInt(galoisFieldSize);          // get a random number.

    return array;              // return the factory built uniform random list.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns an array of integers built using the TrueRNG but
      constrained between the limits of zero and Galois Field cardinality.*/

  public int [] trueRNGUR (int galoisFieldSize)
  {

    int array [] = new int[galoisFieldSize];               // the return field.
    int convertedSample = 0;  // legal values L.T.E. 0.0 & L.T. GF cardinality.

    for (int i = 0; i < galoisFieldSize; i++)
    {
      do
      {
        convertedSample = (int) trueRNG.getNextRandom(); // create real random.
      } while ((convertedSample < 0) || (convertedSample >= galoisFieldSize));
      array[i] = convertedSample;
    }

    return array;              // return the factory built uniform random list.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns an array of integers found using a hybridization of
      the factory RNG reseeded by the TrueRNG hardware device.*/

  public int [] hybridRNG (int galoisFieldSize)
  {

    int array [] = new int[galoisFieldSize];               // the return field.

    for (int i = 0; i < galoisFieldSize; i++)
    {
      RNG.setSeed(trueRNG.getNextRandom());         // re-seed the factory RNG.
      array[i] = RNG.nextInt(galoisFieldSize);          // get a random number.
    }

    return array;              // return the factory built uniform random list.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the horizontal coordinate of a 2-D array where
      the horizontalLimit dimension of the array is known GLOBALLY.*/

  public int findHCoordinate (int uRDist [], int cellNr)
  {

    // all row positions are modulus functions of the width of the matrix.
    return ((uRDist [cellNr]) % this.horizontalLimit);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the vertical coordinate of a 2-D array where
      the horizontalLimit dimension of the array is known GLOBALLY.*/

  public int findVCoordinate (int uRDist [], int cellNr)
  {
    int v = 0;

    // test if value is in the first row, else must be a later row.
    if ( uRDist [cellNr] < this.horizontalLimit)
      v = 0;
    else
      v = uRDist [cellNr] / this.horizontalLimit;

    return v;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and returns the
      name of the biome involved.  It calls a method in terrain to
      accomplish its goals.*/

  public String getBiomeName (int h, int v)
  {

    return terrain.getBiomeName (h, v);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and returns the
      native, average temperature of the biome involved.  It calls a
      method in Climate to accomplish its goals.*/

  public float getBiomeTemp (int h, int v)
  {

    return climate.getBiomeTemp (h, v);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the terrain cell named by the parameters and
      returns the water availability in the biome involved.  It calls a
      method in Terrain to accomplish its goal.*/

  public float getBiomeWater (int h, int v)
  {

    return terrain.getBiomeWater (h, v);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and returns the
      list containing the cell's cohort.  It calls a method in Groups to
      accomplish its goal.*/

  public ArrayList <Hominid> getCohort (int h, int v)
  {

    return groups.getCohort (h, v);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and returns the
      number of hominids in the cell's cohort arraylist.*/

  public int getCohortCount (int h, int v)
  {

    return groups.getCohortCount (h, v);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets this cell location's forage option.  It calls a
      method in Terrain to accomplish its goal.*/

  public String getForageOption (int h, int v)
  {

    return terrain.getForageOption (h, v);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets this cell location's forage (option) probability.
      It calls a method in Terrain to accomplish that goal.*/

  public float getForageProbability (int h, int v)
  {

    return terrain.getForageProbability (h, v);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns pound-equivalent units for a days worth of foraging
      calories.  Foragables include: cassava root, fig, monkey, and termite.
      Calories estimated based on data in "Count Your Calories", Dell Purse
      Book, 1968, Dell Publishing Co., New York: NY. Weight estimates of
      prey (monkey) from Wikipedia on-line.  Termites Name, Date, Title,
      Pub.  Foragables items should match those used in initializeEcology().*/

  public float getForagedMass (String preyName)
  {

    final float FIG_VALUE = 720.0F;  // cal. for 1 pound (16 loose fruit) figs.
    final float FISH_VALUE = 150.0F;            // cal. for  4 oz., small fish.
    final float LEAVES_VALUE = 625.0F; // cal. for  X vol, # lbs. loose leaves.
    final float MONKEY_VALUE = 575.0F;   // cal. for  8 oz. meat, small monkey.
    final float TERMITE_VALUE = 700.0F; //cal. for 125 g., loose live termites.
    final float TUBER_VALUE = 776.0F;        // cal. for  8 small cassava root.

    String input = preyName.trim();   // a cleaned local copy of the parameter.
    float caloricValue = 0.0F;                       // max meal caloric value.

    // match and then compute the equivalent caloric value supplied by each.
    if (input.equalsIgnoreCase("fig"))
      caloricValue = FIG_VALUE * daily_meals;                  // fig calories.
    else if (input.equalsIgnoreCase("leaves"))
       caloricValue = LEAVES_VALUE * daily_meals;           // leaves calories.
    else if (input.equalsIgnoreCase("monkey"))
      caloricValue = MONKEY_VALUE * daily_meals;            // monkey calories.
    else if (input.equalsIgnoreCase("termite"))
      caloricValue = TERMITE_VALUE * daily_meals;          // termite calories.
    else if (input.equalsIgnoreCase("tuber"))
       caloricValue = TUBER_VALUE * daily_meals;             // tuber calories.
    else if (input.equalsIgnoreCase("fish"))
       caloricValue = FISH_VALUE * daily_meals;               // fish calories.
    else if (input.equalsIgnoreCase("none"))
       caloricValue = 0.0F;    // this is the DEFAULT set in Terrain condition.
    else
    {
      System.out.println(input);
      System.out.println("ERROR: Ecology.getForagedResult() bad preyName");
    }

    return caloricValue / FAT_AS_CALORIE;  // a day's worth of foraging in lbs.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and returns the
      elevation of this cell above-below sea level.  It calls a method in
      Terrain to accomplish its goal.*/

  public int getGridElevation (int h, int v)
  {

    return terrain.getElevation (h, v);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets this cell location's current forage probability of
      productivity term. It calls a method in Terrain to accomplish that
      goal. The value is expressed in percent. */

  public float getMaxProductivity (int h, int v)
  {

   return terrain.getMaxProductivity (h, v);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method generates and returns a set of pseudo-random, uniformly
      distributed values.*/

  public int [] getUniformRandom (int galoisFieldSize)
  {

    final byte EMPTY = (byte) 0x00;                        // symbol for EMPTY.
    final byte FULL =  (byte) 0x01;                         // symbol for FULL.

    int [] uniformRandomList = new int[galoisFieldSize];   // the return field.
    byte [] checkList = new byte [galoisFieldSize];       // used for checking.

    for (int z = 0; z < galoisFieldSize; z++)
      checkList [z] = EMPTY;     // fill the backstop list with NOT USED flags.

    int sample = 0;                               // this is the random sample.
    int g = 0;
    do
    { // continue looping until every "g" in the Galois set is tested.

      sample = RNG.nextInt (galoisFieldSize);
      if (checkList [sample] == EMPTY)
      {
        uniformRandomList [g++] = sample;
        checkList [sample] = FULL;
      }

    } while (g < galoisFieldSize);
    checkList = null;                 // return memory to the operating system.

    return uniformRandomList;                // return the uniform random list.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This constrains horizontal movement onto a tordial surface. There are
      three possible states: too far left, too far right, or on grid.*/

  public int hBounds (int possibleH)
  {

    if (possibleH < 0)                                         // too far left.
      return (horizontalLimit + possibleH);        // circle in from the right.

    if (possibleH < horizontalLimit)                 // if less than the limit.
      return possibleH;                            // no problem, stay on grid.

    if (possibleH >= horizontalLimit)                      // if too far right.
      return possibleH - horizontalLimit;           // circle in from the left.

    System.out.println ("ERROR MODE: Ecology.hBounds testing " + possibleH);
    return -1;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method initializes the terrain (GLOBAL grid) subsystem.*/

  // NOTE: Foragables list(s) must match those used in getForagedResult().
  public void initializeEcology ()
  {

    System.out.println("\nCreating artificial socio-spatial ecology.\n");

    // setup terrain, water areas, and the biomes (if they exist) by name.
    grid = terrain.initializeTerrain ();
    System.out.println("Terrain setup completed.");              // DIAGNOSTIC.

    // setup the climate and weather systems sharing the common GLOBAL grid.
    climate = new Climate (this.rng_Seed, grid,
                           this.horizontalLimit, this.verticalLimit);
    grid = climate.initializeClimate (urDistribution);
    System.out.println("Climate setup completed.");              // DIAGNOSTIC.

    // setup the faunal foragables sharing the common GLOBAL grid.
    fauna = new Fauna (this.rng_Seed, grid,
                       this.horizontalLimit, this.verticalLimit);
    grid = fauna.initializeFauna (urDistribution); // setup fauna by elevation.
    System.out.println("Fauna setup completed.");                // DIAGNOSTIC.

    // setup the floral foragables sharing the common GLOBAL grid.
    flora = new Flora (this.rng_Seed, grid,
                       this.horizontalLimit, this.verticalLimit);
    grid = flora.initializeFlora (urDistribution); // setup flora by elevation.
    System.out.println("Flora setup completed.");                // DIAGNOSTIC.

    // setup the agent group systems sharing the common GLOBAL grid.
    groups = new Groups (this.rng_Seed, grid,
                         this.horizontalLimit, this.verticalLimit);
    grid = groups.initializeGroups();
    System.out.println("Groups setup completed.\n");             // DIAGNOSTIC.

    System.out.println("Artificial ecology setup complete.");    // DIAGNOSTIC.
    System.out.println("Terrain cells with >50% production: " +
                        (int) terrainProductivity(50.0F) + "%\n");

    //reportEcology();                                           // DIAGNOSTIC.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method places the hominids into spatially co-located cohorts.
      It examines each member of the population and determines where they
      should be on the grid.  Since this is the first time each individual
      is being associated with a terrain cell, the forage option of the cell
      [the principle foragable in that cell] is assign the respective
      individual as an their intial preferred prey/forageable.*/

  public void initializeGridCohorts (ArrayList <Hominid> population)
  {

    int h = 0;
    int v = 0;
    String realPrey;          // the real prey on the current terrain location.
    Hominid hom = null;                 // a temporary variable for a hominids.
    for (int i = 0; i < population.size(); i++)
    {
      hom = population.get(i);                  // get the first (next) Person.
      h = hom.getCurrentLoc().x;                       // get their X location.
      v = hom.getCurrentLoc().y;                       // get their Y location.
      realPrey = getForageOption (h, v);     // retrieve local, real prey type.
      //System.out.print("Ecology.initializeGridCohort: " + realPrey); // DIAG.
      hom.setPreferredPrey(realPrey);                        // set preference.
      addCohortIndividual (h, v, hom);    // add hominid to cell's cohort list.
    }

    System.out.println("Seed group(s) inserted into ecology.");// announcement.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns as much of the quantity of water requested as is
      currently available in the cell named.  It calls a method in Terrain
      to help accomplish that goal.  Since the volume of water in any
      particular cell on the terrain is directly proportionate to the
      waterbasin amount it "can" have and the number of times it has has been
      drunk from during the current epoch, a reduction of the total available
      volume of the cell's water contents (post consumption by others or
      self occassion) is maintained as a running total.*/

  public float requestWater (int h, int v, float quantity)
  {

    float amountAvailable = 0.0F;         // the amount that "can" be returned.
    float volume = terrain.getBiomeWater (h, v);      // current water in cell.

    if (volume >= quantity)
    {
      terrain.setBiomeWater (h, v, (volume - quantity));    // vol now in cell.
      amountAvailable = quantity;
    }
    else
    {
      terrain.setBiomeWater (h, v, volume);                 // vol now in cell.
    }

    return amountAvailable;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and removes the
      individual hominid (also named by parameter) from the cell's cohort.
      It calls a method in Groups to accomplish its goal.*/

  public void removeCohortIndividual (int h, int v, Object thisHominid)
  {

    groups.removeCohortIndividual (h, v, thisHominid);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and sets the
      elevation of this cell above-below sea level based on the parameter.
      It calls a method in Terrain to accomplish its goal.*/

  public void setGridElevation (int h, int v, int elevation)
  {

    terrain.setElevation (h, v, elevation);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method initializes the grid subsystem.  It instantiates a matrix
      of array cells representing an abstract, 2-D surface.  The surface is
      one cell deep and is distributed across (and completely covers)
      an hLimit x vLimit rectangular grid.*/

  public Cell [][] startTheGrid (Cell [][] grid)
  {

    for (int h = 0; h < this.horizontalLimit; h++)
      for (int v = 0; v < this.verticalLimit; v++)
        grid [h] [v] = new Cell ( this.rng_Seed );          // start each cell.

    return grid;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets this cell location's forage (option) probability.
      It calls a method in Terrain to accomplish that goal.  However, since
      the productivity of any particular cell on the terrain is indirectly
      proportionate to the number of times its product has been recently
      harvested and upon the relative latency of that harvest, a damping
      effect on the residual, available productivity of the cell (post
      harvesting by others or self on some previous occassion) is required.
      An accumulative reduction value, stored in the cell, is applied
      to the forage (option) probability before the result is returned.*/

  public float takeForageable (int h, int v)
  {

    // --------------- PREPARE THE CELL'S LIKELY FORAGEABLE RESULT ------------
    float probOfForageSuccess = 0.0F;       // will become the returned result.
    probOfForageSuccess = terrain.getForageProbability (h, v);     // seasonal.
    float damping = terrain.getMaxProductivity (h, v); // expressed in percent.
    if (damping == 0.0F)
      probOfForageSuccess = 0.0F;            // unfortunately the cell is dead.
    else
    {
      damping *= (damping / 100.0F);                 // cnvrt to 0.0-1.0 scale.
      probOfForageSuccess *= damping;        // the final result after damping.
    }
    // --------------- PREPARE THE CELL'S LIKELY FORAGEABLE RESULT ------------

    // ----------------- REDUCE THE CELL'S CAPACITY TO PRODUCE ----------------
    damping = terrain.getMaxProductivity (h, v);       // expressed in percent.
    if (damping > 0.0F)
    {
      terrain.setMaxProductivity (h, v, (damping - foraging_cost));
    }
    // ----------------- REDUCE THE CELL'S CAPACITY TO PRODUCE ----------------

    return probOfForageSuccess / 100.0F;  // a probability expressed 0.0 - 1.0.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the percentage of the cells in the terrain that
      are currently productive at or above a percentage threshold level.*/

  public float terrainProductivity (float threshold)
  {

    int cellsAboveThreshold = 0;

    for (int h = 0; h < this.horizontalLimit; h++)
    {
      for (int v = 0; v < this.verticalLimit; v++)
      {
        if (terrain.getMaxProductivity (h, v) >= threshold)
          cellsAboveThreshold++;
      }
    }

    return 100 * ( (float) cellsAboveThreshold /
                   (float) terrain.countOfCells () );

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This constrains vertical movement onto a tordial surface. There are
      three possible states: too far above, too far below, or on grid.*/

  public int vBounds (int possibleV)
  {

    if (possibleV < 0)                                        // too far above.
      return verticalLimit + possibleV;           // circle in from the bottom.

    if (possibleV < verticalLimit)                   // if less than the limit.
      return possibleV;                            // no problem, stay on grid.

    if (possibleV >= verticalLimit)                        // if too far below.
      return possibleV - verticalLimit;              // circle in from the top.

    System.out.println ("ERROR MODE: Ecology.vBounds testing " + possibleV);
    return -1;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method reports the entire grid contents.*/

  public void reportEcology ()
  {

    System.out.println("Ecology.report()");
    for (int h = 0; h < this.horizontalLimit; h++)
    {
      for (int v = 0; v < this.verticalLimit; v++)
      {

        System.out.println("....................................");
        System.out.println("For cell h=" + h + ", v=" + v);
        System.out.println("nrInCohort        " +
                           groups.getCohortCount (h, v));
        System.out.println("elevation         " +
                           terrain.getElevation (h, v));
        System.out.println("biomeName         " +
                           terrain.getBiomeName (h, v));
        System.out.println("temperature       " +
                           terrain.getTemperature (h, v));
        System.out.println("water             " +
                           terrain.getBiomeWater (h, v));
        System.out.println("forageOption      " +
                           terrain.getForageOption (h, v));
        System.out.println("forageProbability " +
                           terrain.getForageProbability (h, v));
        System.out.println("forageProbability " +
                           terrain.getForageProbability (h, v));
      }
    }
    System.out.println("....................................");

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method does something.*/

  public void template()
  {

    //add code here.

  } // End method.
  //***************************************************************************

}
