//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      Clans.java                                                       *
//* CLASS    Climate.java. (part of a synthetic ecology)                      *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     07/01/2015                                                       *
//* REVISION X.0.1.4                                                          *
//* BYTES    00,000                                                           *
//* TIME     6:30                                                             *
//* PURPOSE  TBD.                                                             *
//* METHOD   TBD.                                                             *
//* DETAILS  A Java document exists that outlines class operation.  Other     *
//*          documentation in the form of programmer notes apply.             *
//* ENVIRON  Developed on Intel Core-2 Quad Q8200 2.33 GHz CPU, Nvidia-based  *
//*          PCI-X graphics acceleration.  The OS was MS Windows 7 Ultimate.  *
//*          Java IDE was JGrasp version 2.0.0_08.  The Java compiler was     *
//*          version 1.8.0.  Verification tests were run in Java JRE          *
//*          build 1.8.0-b132 with client VM build 14.3-b01.                  *
//*                                                                           *
//*****************************************************************************
//

package root.ecology.climate;                         // sub-directory storage.

import java.util.Random;                                      // Java language.

import root.ModelConstraints;                              // program routines.
import root.ecology.Cell;                                  //            ditto.

//
//*****************************************************************************
// Climate

public class Climate extends ModelConstraints implements Runnable
{

  //
  //***************************************************************************
  // Class constants, variables, and objects.

  // Constants.
  final float MIN_TEMP = 55.0F;      // minimum temperature each cell will get.
  final float RANDOM_TEMP = 10.0F;      // some cells get up to this much more.

  // Variables.
  private long rng_Seed = 0L;                    // local copy of the RNG seed.
  private static int horizontalLimit = 0;           // horizontal grid surface.
  private static int verticalLimit = 0;               // vertical grid surface.
  private int h = 0;                 // a copy of the coordinate into the grid.
  private int v = 0;                 // a copy of the coordinate into the grid.

  private static int [] urDistribution = null;            // a UR distribution.

  // Objects.
  private Random RNG = null;                 // pointer to standard RNG object.
  private Cell grid [][] = null;  // a copy of the pointer to the terrain grid.

  //***************************************************************************
  /** This is a default class constructor.*/

  public Climate ()
  {
  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This is the Climate class constructor.  This class has seasons,
      precipitation, etc.*/

  public Climate (long newSeed, Cell theGrid [][], int hLimit, int vLimit)
  {

    this.rng_Seed = newSeed;                       // copy the GLOBAL RNG seed.
    RNG = new Random(this.rng_Seed);    // instantiate an RNG object with seed.

    this.grid = theGrid;                   // local pointer to the GLOBAL grid.

    horizontalLimit = hLimit;                             // take a local copy.
    verticalLimit = vLimit;                               // take a local copy.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the horizontal coordinate of a 2-D array where
      the horizontalLimit dimension of the array is known GLOBALLY.*/

  public int findHCoordinate (int uRDist [], int cellNr)
  {

    // all row positions are modulus functions of the width of the matrix.
    return ((uRDist [cellNr]) % horizontalLimit);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the vertical coordinate of a 2-D array where
      the horizontalLimit dimension of the array is known GLOBALLY.*/

  public int findVCoordinate (int uRDist [], int cellNr)
  {
    int v = 0;

    // test if value is in the first row, else must be a later row.
    if ( uRDist [cellNr] < horizontalLimit)
      v = 0;
    else
      v = uRDist [cellNr] / horizontalLimit;

    return v;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and returns the
      native, average temperature of the biome involved.*/

  public float getBiomeTemp (int h, int v)
  {

    return this.grid[h][v].temperature;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns a pointer to the local grid.*/

  public synchronized Cell [][] getClimateGrid ()
  {

    return this.grid;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This constrains horizontal movement onto a tordial surface. There are
      three possible states: too far left, too far right, or on grid.*/

  public int hBounds (int possibleH)
  {

    if (possibleH < 0)                                         // too far left.
      return (horizontalLimit + possibleH);        // circle in from the right.

    if (possibleH < horizontalLimit)                 // if less than the limit.
      return possibleH;                            // no problem, stay on grid.

    if (possibleH >= horizontalLimit)                      // if too far right.
      return possibleH - horizontalLimit;           // circle in from the left.

    System.out.println ("ERROR MODE: Climate.hBounds testing " + possibleH);
    return -42;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method initializes the climate subsystem.*/

  public Cell [][]  initializeClimate (int [] urDist)
  {

    urDistribution = urDist;                     // grid cell scanning pattern.

    return updateClimate (urDistribution);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method directs the updating of all climate-related functions.*/

  public void run ()
  {

    this.grid = updateClimate (urDistribution);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and set the
      native, average temperature of the biome involved.*/

  public void setBiomeTemp (int h, int v, float newTemperature)
  {

    this.grid[h][v].temperature = newTemperature;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets a pointer to the local grid.*/

  public synchronized void setClimateGrid (Cell [][] gridPointer)
  {

    this.grid = gridPointer;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  //** This method presets the horizontal coordinate.*/

  public void setHCoord (int newH)
  {

    this.h = newH;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  //** This method presets the vertical coordinate.*/

  public void setVCoord (int newV)
  {

    this.v = newV;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method directs the updating of climate conditions.*/

  public Cell [][]  updateClimate ( int [] urDist )
  {

    final int CELL_COUNT = horizontalLimit * verticalLimit;

    int h = 0;                    // coordinate into the underlying grid array.
    int v = 0;                    // coordinate into the underlying grid array.
    int i = 0;   // counts which is the next cell in the underlying grid array.
    do
    {

      // pick the h and v coordinate of the cell to be initialized
      h = hBounds(findHCoordinate (urDist, i));
      v = vBounds(findVCoordinate (urDist, i));
      setBiomeTemp (h, v, MIN_TEMP + (RNG.nextFloat() * RANDOM_TEMP));

    } while (++i < (CELL_COUNT));

    return this.grid;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This constrains vertical movement onto a tordial surface. There are
      three possible states: too far above, too far below, or on grid.*/

  public int vBounds (int possibleV)
  {

    if (possibleV < 0)                                        // too far above.
      return verticalLimit + possibleV;           // circle in from the bottom.

    if (possibleV < verticalLimit)                   // if less than the limit.
      return possibleV;                            // no problem, stay on grid.

    if (possibleV >= verticalLimit)                        // if too far below.
      return possibleV - verticalLimit;              // circle in from the top.

    System.out.println ("ERROR MODE: Climate.vBounds testing " + possibleV);
    return -42;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method does something.*/

  public void template()
  {

    //add code here.

  } // End method.
  //***************************************************************************

}
