//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      Clans.java                                                       *
//* CLASS    Fauna.java. (part of a synthetic ecology)                        *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     07/01/2015                                                       *
//* REVISION X.0.1.4                                                          *
//* BYTES    00,000                                                           *
//* TIME     6:30                                                             *
//* PURPOSE  To simulate the faunal resources of the simulated ecology.       *
//* METHOD   Multi-threaded elements in object oriented code.                 *
//* DETAILS  A Java document exists that outlines class operation.  Other     *
//*          documentation in the form of programmer notes apply.             *
//* ENVIRON  Developed on Intel Core-2 Quad Q8200 2.33 GHz CPU, Nvidia-based  *
//*          PCI-X graphics acceleration.  The OS was MS Windows 7 Ultimate.  *
//*          Java IDE was JGrasp version 2.0.0_08.  The Java compiler was     *
//*          version 1.8.0.  Verification tests were run in Java JRE          *
//*          build 1.8.0-b132 with client VM build 14.3-b01.                  *
//*                                                                           *
//*****************************************************************************
//

package root.ecology.fauna;                           // sub-directory storage.

import java.util.Random;                                      // Java language.

import root.ecology.Cell;                                  // program routines.
import root.ModelConstraints;                              //            ditto.

//
//*****************************************************************************
// Fauna

public class Fauna extends ModelConstraints implements Runnable
{

  //
  //***************************************************************************
  // Class constants, variables, and objects.

  // Constants.
  // expandable structure holding fauna foragables but *MUST* match the
  // foragables used in Ecology.getForagedResult().
  final String [] FAUNA_FORAGABLES =
                   {"monkey", "monkey",
                    "termite", "termite", "termite", "termite", "termite",
                    "none", "none", "none"}; // 3/10ths are null.

  // Variables.
  private long rng_Seed = 0L;                    // local copy of the RNG seed.
  private int horizontalLimit = 0;                  // horizontal grid surface.
  private int verticalLimit = 0;                      // vertical grid surface.
  private int h = 0;                 // a copy of the coordinate into the grid.
  private int v = 0;                 // a copy of the coordinate into the grid.

  private static int [] urDistribution = null;            // a UR distribution.
  private static int epoch = 0;                      // the current epoch date.

  // Objects.
  private static Random RNG = null;          // pointer to standard RNG object.
  private static Cell grid [][] = null; // copy of pointer to the terrain grid.

  //***************************************************************************
  /** This is a private Class that assists with updating of fauna foraging
      probabilities.*/

  private class UpdateFaunaHelper extends Thread
  {

    private int hStart = 0;
    private int vStart = 0;
    private int hLimit = 0;
    private int vLimit = 0;
    private float sProbability = 0.0F;

    public void setSubgridLimits (int startH, int startV,
                                  int limitH, int limitV) {
      this.hStart = startH;          // the sub-grid horizontal starting point.
      this.vStart = startV;            // the sub-grid vertical starting point.
      this.hLimit = limitH;                   // the sub-grid horizontal limit.
      this.vLimit = limitV;                     // the sub-grid vertical limit.
    }

    public void setSuccessProb (float successProbability) {
      this.sProbability = successProbability;   // prob of successful foraging.
    }

    public void run () {
      float productivity = 0.0F;
      for (int h = this.hStart; h < this.hLimit; h++)
      {
        for (int v = this.vStart; v < this.vLimit; v++)
        {
          productivity = getMaxProductivity (h, v);
          if ((productivity > 0.0F) && (productivity <= 100.0F))
          {
            if (getElevation (h, v) == water)
            {
              // do nothing.
            }
            else if (getElevation (h, v) == lo_land)
            {
              // do nothing.
            }
            else if (getElevation (h, v) == hi_land)
            { // put at h, v, & pntr.
              setBiomeWater (h, v, getBiomeWaterBaseline (h, v));
                setForageProbability (h,  v, this.sProbability);
              if ((productivity + foraging_rcvry_rate) > 100.0F)
                setMaxProductivity (h,  v, 100.0F);
              else
                setMaxProductivity (h,  v,
                                   (productivity + foraging_rcvry_rate));
            }
            else
              System.out.println ("ERROR: Fauna.updateFaunaHelper bad data.");
          }
        }
      }
    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This is the Fauna class constructor.  This class has foragable
      insect and animal species.*/

  public Fauna (long newSeed, Cell theGrid [][], int hLimit, int vLimit)
  {

    this.rng_Seed = newSeed;                       // copy the GLOBAL RNG seed.
    RNG = new Random(this.rng_Seed);    // instantiate an RNG object with seed.

    grid = theGrid;                        // local pointer to the GLOBAL grid.

    this.horizontalLimit = hLimit;                        // take a local copy.
    this.verticalLimit = vLimit;                          // take a local copy.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the horizontal coordinate of a 2-D array where
      the horizontalLimit dimension of the array is known GLOBALLY.*/

  private int findHCoordinate (int uRDist [], int cellNr)
  {

    // all row positions are modulus functions of the width of the matrix.
    return ((uRDist [cellNr]) % this.horizontalLimit);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the vertical coordinate of a 2-D array where
      the horizontalLimit dimension of the array is known GLOBALLY.*/

  private int findVCoordinate (int uRDist [], int cellNr)
  {
    int v = 0;

    // test if value is in the first row, else must be a later row.
    if ( uRDist [cellNr] < this.horizontalLimit)
      v = 0;
    else
      v = uRDist [cellNr] / this.horizontalLimit;

    return v;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and gets the
      name of the biome involved.*/

  private String getBiomeName (int h, int v)
  {

    return grid[h][v].biomeName;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and gets the
      native, water baseline availability of the biome involved.*/

  private float getBiomeWaterBaseline (int h, int v)
  {

    return grid[h][v].waterBaseline;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and returns the
      elevation of this cell above-below sea level.*/

  private int getElevation (int h, int v)
  {

    return grid[h][v].elevation;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns a pointer to the local grid.*/

  public Cell [][] getFaunaGrid ()
  {

    return grid;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets this cell location's forage option.*/

  private String getForageOption (int h, int v)
  {

    return grid[h][v].forageOption;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets this cell location's current forage probability of
      productivity term. This is a value expressed in percent.*/

  private float getMaxProductivity (int h, int v)
  {

    return grid[h][v].maxProductivity;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method initializes the faunal subsystem using randomized order
      and elevation data as a guide..*/

  public Cell [][] initializeFauna ( int [] urDist )
  {

    final float FISH_PROBABILITY = RNG.nextFloat();   // a general probability.
    final float FISH_PRODUCTIVITY = 1.0F * FISH_PROBABILITY;   // specifically.

    urDistribution = urDist;                     // grid cell scanning pattern.

    int pntr = 0;                     // this points into the foragables array.
    int i = 0;                // a cell by cell iterator across the grid array.

    for (int h = 0; h < horizontalLimit; h++)
    {
      for (int v = 0; v < verticalLimit; v++)
      {
        pntr = (urDist[i++] % FAUNA_FORAGABLES.length);    // a Poisson distro.
        if (getElevation (h, v) == water)
        { // put at h, v, & pntr.
          setForageOption (h, v, "fish");
          setForageProbability (h, v, FISH_PROBABILITY);
          setMaxProductivity (h, v, FISH_PRODUCTIVITY);
        }
        else if (getElevation (h, v) == lo_land)
        {
           // do nothing.
        }
        else if (getElevation (h, v) == hi_land)
        { // possibly put at h, v a water basin.
          setMaxProductivity (h, v, initial_productivity);    // initialized.
          setForageOption (h, v, FAUNA_FORAGABLES [pntr]);
          if (!getForageOption (h, v).equals("none"))
          {
            if (unfairCoinToss (rate_of_highland_water_basins))
            {
              float volume = min_water + (RNG.nextFloat() * random_water);
              setBiomeWater (h, v, volume);
              setBiomeWaterBaseline (h, v, volume);
            }
          }
        }
        else
          System.out.println ("ERROR: Fauna.updateFaunaHelper bad data.");
      }
    }

    // the CONSTANT LITERAL 1 represents the first day of the simulation cycle.
    return updateFauna ( 1 );                     // setup foraging parameters.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method directs the updating of all fauna-related functions.*/

  public void run ()
  {

    grid = updateFauna (this.epoch);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and sets the
      name of the biome involved.*/

  private void setBiomeName (int h, int v, String newName)
  {

    grid[h][v].biomeName = newName;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and sets the
      native, water availability of the biome involved.*/

  private void setBiomeWater (int h, int v, float newWater)
  {

    grid[h][v].water = newWater;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and sets the
      native, water baseline availability of the biome involved.*/

  private void setBiomeWaterBaseline (int h, int v, float newWater)
  {

    grid[h][v].waterBaseline = newWater;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the local date.*/

  public void setEpoch (int today)
  {

    this.epoch = today;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets a pointer to the local grid.*/

  public void setFaunaGrid (Cell [][] gridPointer)
  {

    grid = gridPointer;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets this cell location's forage option.*/

  private void setForageOption (int h, int v, String newForageOption)
  {

    grid[h][v].forageOption = newForageOption;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets this cell location's forage (option) probability.*/

  private void setForageProbability (int h, int v,
                                     float newForageProbability)
  {

    grid[h][v].forageProbability = newForageProbability;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  //** This method presets the horizontal coordinate.*/

  private void setHCoord (int newH)
  {

    this.h = newH;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets this cell location's current forage probability of
      productivity term. This is a value expressed in percent.*/

  private void setMaxProductivity (int h, int v, float newProductivity)
  {

    grid[h][v].maxProductivity = newProductivity;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  //** This method presets the vertical coordinate.*/

  private void setVCoord (int newV)
  {

    this.v = newV;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the Boolean result of an unfair coin toss where
      the result of the toss is assured to be 0 (false) if the failureBias
      is 100 or a 1 (true) if the failureBias is 0.  The parameter
      failureBias can be any integer between 0 and 100, inclusive.*/

  private boolean unfairCoinToss (float failureBias)
  {

    if (RNG.nextFloat() <= ((failureBias / 100.0F)))
      return true;
    else
      return false;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method directs the updating of fauna foraging probabilities.*/

  private Cell [][] updateFauna (int epoch)
  {

    // the probability of hunting success is bounded by a floor function.
    float hProbability = RNG.nextFloat ();
    hProbability = ((fauna_success_bias > hProbability) ?
                     fauna_success_bias : hProbability);

    // the sub-grid is divided into tiles 1 sq. km. in size.  The top-level
    // grid is required to be square.  Given that each pixel is defined as a
    // 10 m square, the step across a tile is 100 pixels or 1 km.  And, the
    // number of pixels per tile is 10000.  Thus, by multiplying the sides of
    // the top-level grid and dividing by 10000 we can calculate TILE_COUNT.
    final int STEP = 100;
    final int TILE_PIXELS = 10000;
    final int TILE_COUNT = (this.horizontalLimit * this.verticalLimit) /
                            TILE_PIXELS;
    UpdateFaunaHelper [] thread = new UpdateFaunaHelper [ TILE_COUNT ];

    int tile = 0;
    int i = 0;
    for (int h = 0; h < this.horizontalLimit; h +=STEP)
    {
      for (int v = 0; v < this.verticalLimit; v += STEP)
      {
        //instantiate the threaded object and initialize it.
        thread [ tile ] = new UpdateFaunaHelper ();
        thread [ tile ].setSubgridLimits (h, v, (h + STEP), (v + STEP));
        thread [ tile ].setSuccessProb (hProbability);
        thread [ tile ].start (); // start the thread.
        tile++;
      }
    }

    // loop on the join method until the internal isAlive fails for each.
    for ( int j = 0; j < TILE_COUNT; j++ )
    {
      try
      {
        thread [ j ].join ();
      }
      catch ( InterruptedException e)
      {
        System.out.println ( "ERROR: Fauna.updateFauna" );
      }
    }

    return grid;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method does nothing.*/

  public boolean template ()
  {

    return false;

  } // End method.
  //***************************************************************************

}
