//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      Clans.java                                                       *
//* CLASS    Terrain.java. (a grid-based, n x m synthetic ecology)            *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     07/01/2015                                                       *
//* REVISION X.0.1.4                                                          *
//* BYTES    00,000                                                           *
//* TIME     6:30                                                             *
//* PURPOSE  This object is the binder that holds the grid cells together.    *
//* METHOD   Each instance of a class is a stand-alone code object.           *
//* DETAILS  A Java document exists that outlines class operation.  Other     *
//*          documentation in the form of programmer notes apply.             *
//* ENVIRON  Developed on Intel Core-2 Quad Q8200 2.33 GHz CPU, Nvidia-based  *
//*          PCI-X graphics acceleration.  The OS was MS Windows 7 Ultimate.  *
//*          Java IDE was JGrasp version 2.0.0_08.  The Java compiler was     *
//*          version 1.8.0.  Verification tests were run in Java JRE          *
//*          build 1.8.0-b132 with client VM build 14.3-b01.                  *
//*                                                                           *
//*****************************************************************************
//

package root.ecology.terrain;                         // sub-directory storage.

import java.util.ArrayList;                                   // Java language.
import java.util.Random;                                      //         ditto.

import root.ecology.Cell;                                  // program routines.
import root.ModelConstraints;                              //            ditto.

//
//*****************************************************************************
// Terrain

public class Terrain extends ModelConstraints implements Runnable
{

  //
  //***************************************************************************
  // Class constants, variables, and objects.

  // Constants.
  private float STANDING_BODY_WATER = 1000.0F;  // arbitrary, water vol. (gal).

  // Variables.
  private long rng_Seed = 0L;                    // local copy of the RNG seed.
  private static int horizontalLimit = 0;           // horizontal grid surface.
  private static int verticalLimit = 0;               // vertical grid surface.
  private int h = 0;                 // a copy of the coordinate into the grid.
  private int v = 0;                 // a copy of the coordinate into the grid.

  // Objects.
  private static Random RNG = null;          // pointer to standard RNG object.

  public static Cell grid [][] = null;  // copy of pointer to the terrain grid.

  //***************************************************************************
  /** This is a private Class that assists with updating of flora foraging
      probabilities.*/

  private class UpdateTerrainHelper extends Thread
  {

    private int hStart = 0;
    private int vStart = 0;
    private int hLimit = 0;
    private int vLimit = 0;

    public void setSubgridLimits (int startH, int startV,
                                  int limitH, int limitV) {
      this.hStart = startH;          // the sub-grid horizontal starting point.
      this.vStart = startV;            // the sub-grid vertical starting point.
      this.hLimit = limitH;                   // the sub-grid horizontal limit.
      this.vLimit = limitV;                     // the sub-grid vertical limit.
    }

    public void run () {
      for (int h = this.hStart; h < this.hLimit; h++)
      {
        for (int v = this.vStart; v < this.vLimit; v++)
        {
          if (getElevation (h, v) == water)
          { // put at h, v, & pntr.
            setBiomeWater (h, v, getBiomeWaterBaseline (h, v));
          }
          else if (getElevation (h, v) == lo_land)
          {
            // do nothing.
          }
          else if (getElevation (h, v) == hi_land)
          {
            // do nothing.
          }
          else
            System.out.println ("ERROR: Terrain.updateTerrainHelper " +
                                "badData.");
        }
      }
    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This is the Terrain (grid) class constructor.*/

  public Terrain (long newSeed, Cell theGrid [][], int hLimit, int vLimit)
  {

    this.rng_Seed = newSeed;                       // copy the GLOBAL RNG seed.
    RNG = new Random(this.rng_Seed);    // instantiate an RNG object with seed.

    grid = theGrid;                        // local pointer to the GLOBAL grid.

    this.horizontalLimit = hLimit;                        // take a local copy.
    this.verticalLimit = vLimit;                          // take a local copy.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the percentage of the cells in the terrain that
      are currently productive at or above a percentage threshold level.*/

  public int cellProductivityReport (float threshold)
  {

    int cellsAboveThreshold = 0;

    for (int h = 0; h < horizontalLimit; h++)
    {
      for (int v = 0; v < verticalLimit; v++)
      {
        if (getMaxProductivity (h, v) >= threshold)
          cellsAboveThreshold++;
      }
    }

    return (int) ((float) cellsAboveThreshold / (float) countOfCells ());

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the number of cells in the terrain.*/

  public int countOfCells ()
  {

    return horizontalLimit * verticalLimit;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the horizontal coordinate of a 2-D array where
      the horizontalLimit dimension of the array is known GLOBALLY.*/

  public int findHCoordinate (int uRDist [], int cellNr)
  {

    // all column positions are modulus functions of the width of the matrix.
    return ((uRDist [cellNr]) % horizontalLimit);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the vertical coordinate of a 2-D array where
      the horizontalLimit dimension of the array is known GLOBALLY.*/

  public int findVCoordinate (int uRDist [], int cellNr)
  {
    int v = 0;

    // test if value is in the first row, else must be a later row.
    if ( uRDist [cellNr] < horizontalLimit )
      v = 0;
    else
      v = uRDist [cellNr] / horizontalLimit;

    return v;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and returns the
      name of the biome involved.*/

  public String getBiomeName (int h, int v)
  {

    return grid[h][v].biomeName;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and gets the
      native, water baseline availability of the biome involved.*/

  private float getBiomeWaterBaseline (int h, int v)
  {

    return grid[h][v].waterBaseline;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and returns the
      native, water availability of the biome involved.*/

  public float getBiomeWater (int h, int v)
  {

    return grid[h][v].water;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and returns the
      elevation of this cell above-below sea level.*/

  public int getElevation (int h, int v)
  {

    return grid[h][v].elevation;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets this cell location's forage option.*/

  public String getForageOption (int h, int v)
  {

    return grid[h][v].forageOption;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets this cell location's forage option.*/

  public float getForageProbability (int h, int v)
  {

    return grid[h][v].forageProbability;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets this cell location's current forage probability of
      productivity term. This is a value expressed in percent.*/

  public float getMaxProductivity (int h, int v)
  {

    return grid[h][v].maxProductivity;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and returns the
      cell's temperature.*/

  public float getTemperature (int h, int v)
  {

    return grid[h][v].temperature;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This constrains horizontal movement onto a tordial surface. There are
      three possible states: too far left, too far right, or on grid.*/

  public int hBounds (int possibleH)
  {

    if (possibleH < 0)                                         // too far left.
      return (horizontalLimit + possibleH);        // circle in from the right.

    if (possibleH < horizontalLimit)                 // if less than the limit.
      return possibleH;                            // no problem, stay on grid.

    if (possibleH >= horizontalLimit)                      // if too far right.
      return possibleH - horizontalLimit;           // circle in from the left.

    System.out.println ("ERROR MODE: Terrain.hBounds testing " + possibleH);
    return -1;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method initializes the terrain subsystem (water and biomes) using
      elevation data as a guide.*/

  public Cell [][] initializeTerrain ()
  {

    for (int h = 0; h < horizontalLimit; h++)
    {
      for (int v = 0; v < verticalLimit; v++)
      {
        // based on valued derived from the GUI input of map colors to this...
        if (getElevation (h, v) == water)
        {
          // NOTE: foraging option and probability are set in FAUNA.
          setBiomeName (h, v, "water");                       // this is water.
          setBiomeWater (h, v, STANDING_BODY_WATER); //this is a body of water.
          setBiomeWaterBaseline (h, v, STANDING_BODY_WATER);  // body of water.
        }
        if (getElevation (h, v) == lo_land)
        {
          // NOTE: foraging option and probability are set in FLORA.
          setBiomeName (h, v, "lowland");                      // this is land.
          setBiomeWater (h, v, 0.0F);   // some places might have water basins.
        }
        if (getElevation (h, v) == hi_land)
        {
          // NOTE: foraging option and probability are set in FAUNA.
          setBiomeName (h, v, "highland");                     // this is land.
          setBiomeWater (h, v, 0.0F);   // some places moght have water basins.
        }
      }
    }

    return grid;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method directs the updating of all terrain-related functions.*/

  public void run ()
  {

    grid = updateTerrain ();

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and sets the
      name of the biome involved.*/

  public void setBiomeName (int h, int v, String newName)
  {

    grid[h][v].biomeName = newName;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and sets the
      native, water availability of the biome involved.*/

  public void setBiomeWater (int h, int v, float newWater)
  {

    grid[h][v].water = newWater;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and sets the
      native, water baseline availability of the biome involved.*/

  public void setBiomeWaterBaseline (int h, int v, float newWater)
  {

    grid[h][v].waterBaseline = newWater;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and sets the
      elevation above-below sea level.*/

  public void setElevation (int h, int v, int newElevation)
  {

    grid[h][v].elevation = newElevation;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets this cell location's forage option.*/

  private void setForageOption (int h, int v, String newForageOption)
  {

    grid[h][v].forageOption = newForageOption;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets this cell location's forage (option) probability.*/

  private void setForageProbability (int h, int v, float newForageProbability)
  {

    grid[h][v].forageProbability = newForageProbability;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  //** This method presets the horizontal coordinate.*/

  public void setHCoord (int newH)
  {

    this.h = newH;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets this cell location's current forage probability of
      productivity term. This is a value expressed in percent.*/

  public void setMaxProductivity (int h, int v, float newProductivity)
  {

    grid[h][v].maxProductivity = newProductivity;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets a pointer to the local grid.*/

  public void setTerrainGrid (Cell [][] gridPointer)
  {

    grid = gridPointer;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  //** This method presets the vertical coordinate.*/

  public void setVCoord (int newV)
  {

    this.v = newV;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method directs the updating of selected terrain characteristics.*/

  private Cell [][] updateTerrain ()
  {

    // the sub-grid is divided into tiles 1 sq. km. in size.  The top-level
    // grid is required to be square.  Given that each pixel is defined as a
    // 10 m square, the step across a tile is 100 pixels or 1 km.  And, the
    // number of pixels per tile is 10000.  Thus, by multiplying the sides of
    // the top-level grid and dividing by 10000 we can calculate TILE_COUNT.
    final int STEP = 100;
    final int TILE_PIXELS = 10000;
    final int TILE_COUNT = (this.horizontalLimit * this.verticalLimit) /
                            TILE_PIXELS;
    UpdateTerrainHelper [] thread = new UpdateTerrainHelper [ TILE_COUNT ];

    int tile = 0;
    for (int h = 0; h < this.horizontalLimit; h +=STEP)
    {
      for (int v = 0; v < this.verticalLimit; v += STEP)
      {
        //instantiate the threaded object and initialize it.
        thread [ tile ] = new UpdateTerrainHelper ();
        thread [ tile ].setSubgridLimits (h, v, (h + STEP), (v + STEP));
        thread [ tile ].start (); // start the thread.
        tile++;
      }
    }

    // loop on the join method until the internal isAlive fails for each.
    for ( int i = 0; i < TILE_COUNT; i++ )
    {
      try
      {
        thread [ i ].join ();
      }
      catch ( InterruptedException e)
      {
        System.out.println ( "ERROR: Terrain.updateTerrain" );
      }
    }

    return grid;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This constrains vertical movement onto a tordial surface. There are
      three possible states: too far above, too far below, or on grid.*/

  public int vBounds (int possibleV)
  {

    if (possibleV < 0)                                        // too far above.
      return verticalLimit + possibleV;           // circle in from the bottom.

    if (possibleV < verticalLimit)                   // if less than the limit.
      return possibleV;                            // no problem, stay on grid.

    if (possibleV >= verticalLimit)                        // if too far below.
      return possibleV - verticalLimit;              // circle in from the top.

    System.out.println ("ERROR MODE: Terrain.vBounds testing " + possibleV);
    return -1;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method does nothing.*/

  public boolean template ()
  {

    return false;

  } // End method.
  //***************************************************************************

}
